import copy

from datapackage_pipelines.wrapper import ingest, spew
from datapackage_pipelines.utilities.kvstore import DB


def saver(resource, db):
    for idx, row in enumerate(resource):
        key = "{:08x}".format(idx)
        db.set(key, row)
        yield row


def loader(db):
    for k, value in db.items():
        yield value


def process_resources(resource_iterator, source):
    for resource in resource_iterator:
        if resource.spec['name'] == source:
            db = DB()
            yield saver(resource, db)
            yield loader(db)
        else:
            yield resource


def process_datapackage(dp, source, target_name, target_path):

    def traverse_resources(resources):
        for res in resources:
            yield res
            if res['name'] == source:
                res = copy.deepcopy(res)
                res['name'] = target_name
                res['path'] = target_path
                yield res

    dp['resources'] = list(traverse_resources(dp['resources']))
    return dp


if __name__ == '__main__':
    parameters, datapackage, resource_iterator = ingest()

    source = parameters['source']
    target_name = parameters['target-name']
    target_path = parameters['target-path']

    spew(process_datapackage(datapackage, source, target_name, target_path),
         process_resources(resource_iterator, source))



import copy
import re

from datapackage_pipelines.wrapper import ingest, spew
from datapackage_pipelines.utilities.resource_matcher import ResourceMatcher

parameters, datapackage, resource_iterator = ingest()

resources = ResourceMatcher(parameters.get('resources'))
unpivot_fields = parameters.get('unpivot')
extra_keys = parameters.get('extraKeyFields')
extra_value = parameters.get('extraValueField')


def match_fields(field_name_re, expected):
    def filt(field):
        return (field_name_re.fullmatch(field['name']) is not None) is expected
    return filt


def process_datapackage(datapackage_):
    unpivot_fields_without_regex = []
    for resource in datapackage_['resources']:
        name = resource['name']
        if not resources.match(name):
            continue

        if 'schema' not in resource:
            continue

        fields = resource['schema'].get('fields', [])

        for u_field in unpivot_fields:
            field_name_re = re.compile(u_field['name'])
            fields_to_pivot = (list(
                filter(match_fields(field_name_re, True), fields)
            ))
            fields = list(
                filter(match_fields(field_name_re, False), fields)
            )

            # handle with regex
            for field_to_pivot in fields_to_pivot:
                original_key_values = u_field['keys']  # With regex
                new_key_values = {}
                for key in original_key_values:
                    new_val = original_key_values[key]
                    if isinstance(new_val, str):
                        new_val = re.sub(
                            u_field['name'], new_val, field_to_pivot['name'])
                    new_key_values[key] = new_val
                    field_to_pivot['keys'] = new_key_values
                unpivot_fields_without_regex.append(field_to_pivot)

        fields_to_keep = [f['name'] for f in fields]
        fields.extend(extra_keys)
        fields.append(extra_value)
        resource['schema']['fields'] = fields
    return unpivot_fields_without_regex, fields_to_keep


def unpivot(rows, fields_to_unpivot_, fields_to_keep_):
    for row in rows:
        for unpivot_field in fields_to_unpivot_:
            new_row = copy.deepcopy(unpivot_field['keys'])
            for field in fields_to_keep_:
                new_row[field] = row[field]
            new_row[extra_value['name']] = row.get(unpivot_field['name'])
            yield new_row


def process_resources(resource_iterator_, fields_to_unpivot, fields_to_keep):
    for resource in resource_iterator_:
        spec = resource.spec
        if not resources.match(spec['name']):
            yield resource
        else:
            yield unpivot(resource, fields_to_unpivot, fields_to_keep)


old_fields, keep_fields = process_datapackage(datapackage)

spew(datapackage, process_resources(resource_iterator, old_fields, keep_fields))



import re

from datapackage_pipelines.wrapper import ingest, spew
from datapackage_pipelines.utilities.resource_matcher import ResourceMatcher

parameters, datapackage, resource_iterator = ingest()

resources = ResourceMatcher(parameters.get('resources'))
fields = parameters.get('fields', [])


def process_resource(rows):
    for row in rows:
        for field in fields:
            for pattern in field.get('patterns', []):
                row[field['name']] = re.sub(
                    str(pattern['find']),
                    str(pattern['replace']),
                    str(row[field['name']]))
        yield row


def process_resources(resource_iterator_):
    for resource in resource_iterator_:
        spec = resource.spec
        if not resources.match(spec['name']):
            yield resource
        else:
            yield process_resource(resource)


spew(datapackage, process_resources(resource_iterator))



import functools
import collections

from datapackage_pipelines.wrapper import ingest, spew
from datapackage_pipelines.utilities.resource_matcher import ResourceMatcher

parameters, datapackage, resource_iterator = ingest()

resources = ResourceMatcher(parameters.get('resources'))
fields = parameters.get('fields', [])


def modify_datapackage(datapackage_):
    dp_resources = datapackage_.get('resources', [])
    for resource_ in dp_resources:
        if resources.match(resource_['name']):
            new_fields = [
                {
                    'name': f['target'],
                    'type': get_type(resource_['schema']['fields'],
                                    f.get('source', []),
                                    f['operation'])
                } for f in fields
            ]
            resource_['schema']['fields'] += new_fields
    return datapackage_


def get_type(res_fields, operation_fields, operation):
    types = [f.get('type') for f in res_fields if f['name'] in operation_fields]
    if 'any' in types:
        return 'any'
    if (operation == 'format') or (operation == 'join'):
        return 'string'
    if ('number' in types) or (operation == 'avg'):
        return 'number'
    # integers
    if len(types):
        return types[0]
    # constant
    return 'any'


def process_resource(rows):
    for row in rows:
        for field in fields:
            values = [
                row.get(c) for c in field.get('source', []) if row.get(c) is not None
            ]
            with_ = field.get('with', '')
            new_col = AGGREGATORS[field['operation']].func(values, with_, row)
            row[field['target']] = new_col
        yield row


def process_resources(resource_iterator_):
    for resource in resource_iterator_:
        spec = resource.spec
        if not resources.match(spec['name']):
            yield resource
        else:
            yield process_resource(resource)


Aggregator = collections.namedtuple('Aggregator', ['func'])

AGGREGATORS = {
    'sum': Aggregator(lambda values, fstr, row: sum(values)),
    'avg': Aggregator(lambda values, fstr, row: sum(values) / len(values)),
    'max': Aggregator(lambda values, fstr, row: max(values)),
    'min': Aggregator(lambda values, fstr, row: min(values)),
    'multiply': Aggregator(
        lambda values, fstr, row: functools.reduce(lambda x, y: x*y, values)),
    'constant': Aggregator(lambda values, fstr, row: fstr),
    'join': Aggregator(
        lambda values, fstr, row: fstr.join([str(x) for x in values])),
    'format': Aggregator(lambda values, fstr, row: fstr.format(**row)),
}


spew(modify_datapackage(datapackage), process_resources(resource_iterator))
